@@ -5,7 +5,7 @@ use strict;
 use warnings;
 
 use vars qw($VERSION);
-$VERSION = '0.16';
+$VERSION = '0.23';
 
 # The following are for debugging only
 #use ExtUtils::Embed;
@@ -13,7 +13,7 @@ $VERSION = '0.16';
 #use Inline Config => CLEAN_AFTER_BUILD => 0; # cp _Inline/build/Text/Scan/Scan.xs .
 
 use Inline C => 'DATA',
-			VERSION => '0.16',
+			VERSION => '0.23',
 			NAME => 'Text::Scan';
 
 sub serialize {
@@ -26,6 +26,7 @@ sub restore {
 	return _restore( $self, "$filename.trie", "$filename.vals");
 }
 
+
 1;
 
 __DATA__
@@ -76,20 +77,32 @@ Text::Scan - Fast search for very large numbers of keys in a body of text.
 	# (new in v0.10)
 	@vals = $dict->values();
 	# @vals is ( canine, ursine, porcine )
-	
+
+	# Get back everything you inserted
+	%everything = $dict->dump();
+
 	# "mindex"
 	# Like perl's index() but with multiple patterns (new in v0.07)
 	# you can scan for the starting positions of terms.
 	@indices = $dict->mindex( $document );
 	# @indices is ( dog => 4, bear => 16, dog => 29 ) 
 
-	# The hash context yeilds the position of the last occurrences 
+	# The hash context yields the position of the last occurrences 
 	# of each word 
 	%indices = $dict->mindex( $document ); 
 	# %indices is ( dog => 26, bear => 16 )
 
-	# Turn on wildcard scanning. (New in v0.09) 
-	# This can be done anytime. Works for scan() and mindex()
+	# multiscan() (>= v0.23)
+	# Retrieves everything scan() and mindex() does, in the form
+	# of an array of references. Each reference points to a list
+	# of (key, index, value)
+	@result = $dict->multiscan($document);
+	# @result is ([dog, 4, canine], [bear, 16, ursine], [dog, 29, canine])
+
+
+	# Turn on wildcard scanning. (>= v0.09) 
+	# This can be done anytime. Works for scan() and mindex(). Wildcards
+	# encompass any number of non-single-space-equivalent chars.
 	$dict->usewild();
 
 	# Save a dictionary, then restore it. (serialize and restore new in v0.14)
@@ -99,21 +112,52 @@ Text::Scan - Fast search for very large numbers of keys in a body of text.
 	$dict->serialize("dict_name");
 	$dict->restore("dict_name");
 
-	
+	# Place a global char equivalency class into effect. This matches all
+	# these characters as if they were the same. (v0.17)
+	$dict->charclass(".:;,?");
+	$dict->insert("What?", "What?");
+	@found = $dict->scan("Err... What, something wrong?");
+	# now @found is ( "What," => "What?" );
+
+	# Scan case-insensitively. This must be called before any insertions.
+	$dict->ignorecase();
+
+	# Set a class of chars to be the boundaries of any match, 
+	# such that the chars immediately before the beginning and after the
+	# ending of a match have to be in this class. Default is the single
+	# space. (beginning and ending of string always count as bounds)
+	# This can be called at any time, and supercedes any previous calls.
+	$dict->boundary(".? ");
+
+	# Ignore certain chars. You can define a class of chars that the
+	# dictionary should pretend do not exist. You must call this before
+	# any insertions.
+	$dict->ignore("\n\r<>()");
+
+
 =head1 DESCRIPTION
 
-This module provides facilities for fast searching on arbitrarily long texts with very many search keys. The basic object behaves somewhat like a perl hash, except that you can retrieve based on a superstring of any keys stored. Simply scan a string as shown above and you will get back a perl hash (or list) of all keys found in the string (along with associated values (or positions if you use mindex() instead of scan(), see examples above)). All keys present in the text are returned.
+This module provides facilities for fast searching on strings with very many search keys. The basic object behaves somewhat like a perl hash, except that you can retrieve based on a superstring of any keys stored. Simply scan a string as shown above and you will get back a perl hash (or list) of all keys found in the string (along with associated values and/or positions). All keys present in the text are returned.
 
-NOTE: This is a behavioral change from previous versions where keys could never overlap. Now they may overlap and still be detected.
+There are several ways to influence the behavior of the match, chiefly by the use of several types of B<global character classes>. These are different from regular expression char classes, in that they apply to the entire text and for all keys. These consist of the "ignore" class, the "boundary" class, and any user-defined classes.
 
-IMPORTANT: A B<single space> is used as a delimiter for purposes of recognizing key boundaries. That's right, there is a bias in favor of processing natural language! In other words, if 'my dog' is a key and 'my dogs bite' is the text, 'my dog' will B<not> be recognized. I plan to make this more configurable in the future, to have a different delimiter or none at all. For now, recognize that the key 'drunk' will not be found in the text 'gedrunk' or 'drunken' (or 'drunk.' for that matter). Properly tokenizing your corpus is essential. I know there is probably a better solution to the problem of substrings, and if anyone has suggestions, by all means contact me.
+Using "ignore" characters you can have the scan pretend a char in the text simply does not exist. This is useful if you want to avoid tokenizing your text. So for instance, if the period '.' is in your "ignore" class, the text will be treated exactly as if all periods had been deleted.
 
-=head1 COMMENTARY
+To define what characters may count as the delimiter of any match (single space by default) you can use the "boundary" class. For instance this way you can count punctuation as a boundary, and phrases bounded at the end by punctuation will match.
+
+Any user-defined character classes can be used to count different chars as the same. For instance this is used internally to implement case-insensitive matching.
 
-What I am leaning toward is simply having no implicit delimiter at all, and relying on the programmer to use a chosen delimiter when inserting keys, then tokenizing the target text properly so that the delimiter is present at boundaries as defined by your application. This would leave you free to have no delimiter if you really want "drunk" to match "gedrunk", "drunken", "drunk." etc. The chore of tokenizing the target would be mitigated by pattern matching capabilities (hmm..)
 
 =head1 NEW
 
+In v 0.19: "boundary" character class defines legal boundary characters for all matches. Default is single space for backward compatibility.
+
+In v 0.18: Global "ignore" character classes. This, along with general global char classes and case-insensitivity, should allow you to eliminate most preprocessing.
+
+In v 0.17: Global character classes, see example above. Also thereby case-insensitivity.
+
+In v 0.16: Now all patterns present in the text are returned regardless of where they begin or end.
+
 In v 0.13: A more-or-less complete rewrite of Text::Scan uses a more traditional finite-state machine rather than a ternary trie for the basic data structure. This results in an average 20% savings in memory and 10% savings in runtime, besides being much simpler to implement, thus less prone to bugs.
 
 In v 0.09: Wildcards! A limited wildcard functionality is available. call usewild() to turn it on. Thereafter any asterisk (*) followed by a space (' ') will be treated as "zero or more non-space characters". Once this function is turned on, the scan will be approximately 50% slower than with literal strings. If you include '*' in any key without calling usewild(), the '*' will be treated literally.
@@ -165,7 +209,7 @@ This library is free software; you can redistribute it and/or modify it under th
 
 =head1 AUTHOR
 
-Ira Woodhead, ira at foobox dot com
+Ira Woodhead, textscan at sweetpota dot to
 
 =cut
 
@@ -189,25 +233,44 @@ typedef struct FSM {
 	int transitions;
 	int states;
 	int maxpath;
+	char* ignore;
+	char* boundary;
+	char* charclasses;
+	char* wild;     // all chars which match a wildcard
 	AV* found_keys;
+	AV* found_offsets;
 	AV* found_vals;
-	bool use_wildcards;
+	int position;
+    bool use_wildcards;
+	char* s;      // string on which to match
 } Fsm;
 
 
-// A link in a list of pending matches, for use with wildcards.
-typedef struct pmatch *pmPtr;
-typedef struct pmatch { 
-	int depth; 
-	pmPtr next;
-	trans p;
-	char *t;
-	bool inWild;
-} pMatch;
+// function declarations (prototypes) necessary for 
+// mutually recursive functions
+int _eat_wild_chars(fsm this, int matchlen, trans p);
+int _find_match(fsm this, int matchlen, trans p);
+
+
+// For vector records
+#define BIT_ON(vec, offset, pos) \
+		( *(vec + \
+			((unsigned char) offset*32) + \
+			(unsigned char)pos/8) |= \
+			(1 << ((unsigned char)pos % 8)) \
+		)
+
+#define IS_BIT_ON(vec, offset, pos) \
+		( *(vec+((unsigned char) offset*32) + \
+			(unsigned char)pos/8) & \
+			(1 << ((unsigned char)pos % 8)) \
+		)
+
 
 
 _malloc(fsm m) {
 	av_clear(m->found_keys);
+	av_clear(m->found_offsets);
 	av_clear(m->found_vals);
 }
 
@@ -247,7 +310,8 @@ trans _insert_(fsm m, trans p, char *s, SV* val) {
 	
 	// search for *s in transition list (state) t
 	while(t){
-		if(*s == t->splitchar) break;
+//		if(*s == t->splitchar) break;
+		if(IS_BIT_ON(m->charclasses, *s, t->splitchar)) break;
 		else t = t->next_trans;
 	}
 
@@ -279,7 +343,8 @@ trans _insert_(fsm m, trans p, char *s, SV* val) {
 }
 
 
-
+// unused, too slow
+/*
 void _cleanup_(trans p) {
 
 	if (p) {
@@ -294,6 +359,7 @@ void _cleanup_(trans p) {
 	}
 	//free(p);
 }
+*/
 
 
 int _search(trans root, char *s) {
@@ -313,73 +379,74 @@ int _search(trans root, char *s) {
 
 
 // Return the node representing the char s, if it exists, from this list.
-trans _bsearch( trans q, char s ){
+trans _bsearch( char* vec, char s, trans q ){
 
-//	trans parent = q;
 	while(q){
-		if(s == q->splitchar) break;
-		else {
-//			parent = q;
+		if(IS_BIT_ON(vec, s, q->splitchar)) 
+			break;
+		else 
 			q = q->next_trans;
-		}
 	}
-//	_rotate(parent);
 	return q;
 }
 
-void _record_match(char *s,
-					int position,
-					int matchlen, 
-					AV* keys, 
-					AV* vals, 
-					SV* val, 
-					bool isMindex){
+void _record_match(fsm this, int matchlen, trans p){
 
-	av_push(keys, newSVpvn(s,matchlen+1));
-
-	if(isMindex){
-		av_push(vals, newSViv(position));
-	}
-	else {
-		av_push(vals, val);
-		SvREFCNT_inc(val);
-	}
+	SV* val = (SV*) p->next_state;
+	av_push(this->found_keys,    newSVpvn(this->s,matchlen+1));
+	av_push(this->found_offsets, newSViv(this->position));
+	av_push(this->found_vals,    val);
+	SvREFCNT_inc(val);
 }
 
+int _eat_wild_chars(fsm this, int matchlen, trans p){
+	char* t = this->s + matchlen;
+	
+	while( IS_BIT_ON( this->wild, 0, *t ) ){
+		t++;
+		matchlen++;
+	}
+	
+	p = p->next_state;
+	return _find_match(this, matchlen, p);
+}
 
-int _find_literal_match(trans root, 
-						char *s, 
-						int position, 
-						AV* keys, 
-						AV* vals, 
-						bool isMindex){
+int _find_match(fsm this, int matchlen, trans p){
 
-	int matchlen = 0;
-	int depth = 0;
+// These items are invariant through a complete recursive call.
+//this->s             (document to match)
+//this->position      (position in document where s starts)
+//this->found_keys    (perl list of found text)
+//this->found_offsets (perl list of offsets for found text)
+//this->found_vals    (perl list of found values stored in fsm)
 
-	trans p = root;
-	char* t = s;
+	int depth = matchlen;
+	char* t = this->s + matchlen; //starting point for this match
 
 	while(p){
 
 		// if this is a termination state
 		if(!p->splitchar){
-			if( *t == ' ' || *t == 0 ){
-				_record_match(s,
-							position,
-							depth-1, 
-							keys, 
-							vals, 
-							(SV*) p->next_state, 
-							isMindex);
-				//*champaddr = (SV*) p->next_state;
+			if( IS_BIT_ON( this->boundary, 0, *t ) ){
 				matchlen = depth - 1;
+				_record_match(this, matchlen, p);
 			}
 			p = p->next_trans;
 		}
 
+		// ignore irrelevant chars
+		while( IS_BIT_ON(this->ignore, 0, *t) ){
+			t++;
+			depth += 1;
+		}
+
+		// find wildcard matches
+		if(p && p->splitchar == '*' && this->use_wildcards)
+			matchlen = _eat_wild_chars(this, depth, p);
+
+
 		// search for t
-		p = _bsearch( p, *t );
+		p = _bsearch( this->charclasses, *t, p );
 
 		if(p){
 			t++;
@@ -392,199 +459,76 @@ int _find_literal_match(trans root,
 }
 
 
+void _scan(fsm this, char *s) {
 
-
-
-int _find_wild_match(trans root, 
-					char *s, 
-					int position, 
-					AV* keys, 
-					AV* vals, 
-					bool isMindex){
-
-	trans wildp;
-	pmPtr tm;
-	pmPtr temptm;
-	char wild = '*';
-//	char* wp = &wild;
-	
-	int matchlen = 0;
-	
-	tm  = (pmPtr) malloc(sizeof(pMatch));
-	tm->depth = 0;
-	tm->next = 0;
-	tm->p = root;
-	tm->t = s;
-	tm->inWild = FALSE;
-
-	// successful match in progress, longest
-	// match stored in "champaddr".
-	while(tm && tm->p){
-		if( tm->inWild ){
-			if( *(tm->t) == ' ' || *(tm->t) == 0 ){
-				// record a match if possible
-				if( !tm->p->splitchar && tm->depth > matchlen ){
-					_record_match(s,
-								position,
-								tm->depth - 1, 
-								keys, 
-								vals, 
-								(SV*) tm->p->next_state, 
-								isMindex);
-					//*champaddr = (SV*) tm->p->next_state;
-					matchlen = tm->depth - 1;
-				}
-				// advance p out of wildcard, loop with same char ' '
-				tm->p = tm->p->next_state;
-				tm->inWild = FALSE;
-			}
-			else {
-				// eat next char, stay on same p
-				tm->t++;
-				tm->depth++;
-			}
-		}
-		else { // Not on a wildcard
-
-			if( !tm->p->splitchar ){
-				if( tm->depth > matchlen && 
-					(*(tm->t) == ' ' || *(tm->t) == 0) ){
-					_record_match(s,
-								position,
-								tm->depth - 1, 
-								keys, 
-								vals, 
-								(SV*) tm->p->next_state, 
-								isMindex);
-					//*champaddr = (SV*) tm->p->next_state;
-					matchlen = tm->depth - 1;
-				}
-				tm->p = tm->p->next_trans;
-			}
-
-			// Check for a wildcards (There's got to be a better way!)
-			// Insert new match state in next pos in linked list.
-			if( tm->p && tm->p->splitchar == wild ){ //Branch off 
-				temptm = (pmPtr) malloc(sizeof(pMatch));
-				temptm->p = tm->p;
-				temptm->t = tm->t;
-				temptm->depth = tm->depth;
-				temptm->next = tm->next;
-				temptm->inWild = TRUE;
-				tm->next = temptm;
-
-				tm->p = tm->p->next_trans;
-			}
-
-			// search for t, increment p
-			tm->p = _bsearch( tm->p, *tm->t );
-
-			if(tm->p){
-				tm->t++;
-				tm->depth++;
-				tm->p = tm->p->next_state;
-			}
-		}
-		// fall back on previous match state, if any
-		if(!tm->p){
-			temptm = tm->next;
-			free(tm);
-			tm = temptm; 
-		} //Go back to last branch
-	}
-
-	while(tm){
-		temptm = tm->next;
-		free(tm);
-		tm = temptm;
-	}
-
-	return matchlen;
-}
-
-
-
-void _scan(fsm m, trans root, char *s, bool isMindex) {
-
-	AV* keys = m->found_keys;
-	AV* vals = m->found_vals;
-//	SV* champ = 0;
 	char* t;
-	int matchlen = 0;
+	int match = 0;
 	int position = 0;
-
 	
 	while(*s){
-		
-		if(m->use_wildcards){
-			matchlen = _find_wild_match(root, s, position, keys, vals, isMindex);
-		}
-		else {
-			matchlen = _find_literal_match(root, s, position, keys, vals, isMindex);
-		}
+		this->s = s;
+		this->position = position;
+		match = _find_match(this, 0, this->root); 
+
 		// truncate s by length of match or first word...
-		if(matchlen){
+		if(match){ s++; position++; }
 
-/*			av_push(keys, newSVpvn(s,matchlen+1));
+		//Move to the first boundary char
+		while( ! IS_BIT_ON(this->boundary, 0, *s) ) { s++; position++; }
 
-			if(isMindex){
-				av_push(vals, newSViv(position));
-			}
-			else {
-				av_push(vals, champ);
-				SvREFCNT_inc(champ);
-			}
-*/
+		// chop off the first boundary
+		if(*s != 0) { s++; position++; }
 
-// The following substitution makes the scan start at every word,
-// not just after the last match.
-			s++;
-			position++;
-//			s += matchlen;
-//			position += matchlen;
-		}
-
-		while( (*s != ' ') && (*s != 0) ) { s++; position++; }
-
-		if(*s != 0) { s++; position++; } // chop off the space
-		matchlen = 0;
+		// move past any irrelevant chars
+		while( IS_BIT_ON(this->ignore, 0, *s) ) { s++; position++; }
+		match = 0;
 	}
-
 }
 
 
 
-
-void _keys(fsm m, trans p, char* k, int depth) {
+void _dump(fsm m, trans p, char* k, int depth) {
   
 	if (!p) return;
 
-	_keys(m, p->next_trans, k, depth);
+	_dump(m, p->next_trans, k, depth);
 
 	if (p->splitchar){
 		*(k+depth) = p->splitchar;
-		_keys(m, p->next_state, k, depth+1);
+		_dump(m, p->next_state, k, depth+1);
 	}
-	else
-		av_push(m->found_keys, newSVpvn(k, depth));
-}
-
-
-void _values(fsm m, trans p){
-
-	if (!p) return;
-
-	_values(m, p->next_trans);
-	
-	if(p->splitchar)
-		_values(m, p->next_state);
 	else {
-		av_push(m->found_keys, (SV*)p->next_state);
+		av_push(m->found_keys, newSVpvn(k, depth));
+		av_push(m->found_vals, (SV*)p->next_state);
 		SvREFCNT_inc((SV*)p->next_state);
 	}
 }
 
 
+void _init_charclasses(char* vecs){
+	int i;
+	for(i=0;i<256;i++){
+		// For the ith 256-bit span, turn on bit i
+		BIT_ON( vecs, i, i );
+	}
+}
+
+// Default pattern boundary is EOS (null) and space (' ')
+void _init_boundary(char* vec){
+	BIT_ON( vec, 0, 0 );
+	BIT_ON( vec, 0, (int) ' ' );
+}
+
+void _init_wild(char* vec){
+	int i;
+	for(i=1;i<256;i++){
+		// All non-space chars match wilds by default
+		if( !isspace(i) )
+			BIT_ON( vec, 0, i );
+	}
+
+}
+
 SV* new(char* class){
 	fsm m = (fsm) malloc( sizeof(Fsm) );
 	SV* obj_ref = newSViv(0);
@@ -596,7 +540,17 @@ SV* new(char* class){
 	m->states = 0;  
 	m->maxpath = 0;
 
+	m->ignore   = (char*) calloc(256/sizeof(char), sizeof(char));
+	m->boundary = (char*) calloc(256/sizeof(char), sizeof(char));
+	m->wild     = (char*) calloc(256/sizeof(char), sizeof(char));
+	m->charclasses = (char*) calloc((256*256)/sizeof(char), sizeof(char));
+
+	_init_boundary(m->boundary);
+	_init_charclasses(m->charclasses);
+	_init_wild(m->wild);
+
 	m->found_keys = (AV*) newAV(); 
+	m->found_offsets = (AV*) newAV();
 	m->found_vals = (AV*) newAV();
 
 	m->use_wildcards = FALSE;
@@ -611,7 +565,9 @@ void DESTROY(SV* obj){
 
 // takes *far* too long compared to OS garbage collection.
 //	_cleanup_(m->root);
-	free(m);
+//	free(m->charclasses);
+//	free(m->ignore);
+//	free(m);
 }
 
 void usewild(SV* obj){
@@ -620,6 +576,94 @@ void usewild(SV* obj){
 }
 
 
+// This must be called before any insert(), but may be called
+// any number of times.
+void charclass(SV* obj, char* vecstring){
+	fsm m = (fsm)SvIV(SvRV(obj));
+	char* i = vecstring;
+	char* j = vecstring;
+	char* vec = m->charclasses;
+	while(*i){
+		while(*j){
+			// For the ith 256-bit span, turn on the jth bit
+			BIT_ON( vec, *i, *j);
+			j++;
+		}
+		j = vecstring;
+		i++;
+	}
+}
+
+void ignore(SV* obj, char* vecstring){
+	fsm m = (fsm)SvIV(SvRV(obj));
+	char* i = vecstring;
+	for(; *i; i++ )
+		BIT_ON( m->ignore, 0, *i);
+
+	// "ignore" chars also count as boundaries
+	i = vecstring;
+	for(; *i; i++ )
+		BIT_ON( m->boundary, 0, *i );
+
+	// "ignore" chars also match wildcards
+	i = vecstring;
+	for(; *i; i++ )
+		BIT_ON( m->wild, 0, *i );
+}
+
+void ignorecase(SV* obj){
+	charclass(obj, "Aa");
+	charclass(obj, "Bb");
+	charclass(obj, "Cc");
+	charclass(obj, "Dd");
+	charclass(obj, "Ee");
+	charclass(obj, "Ff");
+	charclass(obj, "Gg");
+	charclass(obj, "Hh");
+	charclass(obj, "Ii");
+	charclass(obj, "Jj");
+	charclass(obj, "Kk");
+	charclass(obj, "Ll");
+	charclass(obj, "Mm");
+	charclass(obj, "Nn");
+	charclass(obj, "Oo");
+	charclass(obj, "Pp");
+	charclass(obj, "Qq");
+	charclass(obj, "Rr");
+	charclass(obj, "Ss");
+	charclass(obj, "Tt");
+	charclass(obj, "Uu");
+	charclass(obj, "Vv");
+	charclass(obj, "Ww");
+	charclass(obj, "Xx");
+	charclass(obj, "Yy");
+	charclass(obj, "Zz");
+}
+
+// define class of chars that qualify as beginning/ending of patterns
+// meaning, what is allowed to occur right before or after pattern
+void boundary(SV* obj, char* b){
+	fsm m = (fsm)SvIV(SvRV(obj));
+	int i;
+
+	// Reset boundary to none
+	for( i=0; i<(256/sizeof(char)); i++ )
+		*(m->boundary + i) = 0;
+
+	BIT_ON( m->boundary, 0, 0 );
+
+	// Special case: if null string specified, all chars are boundary
+	if(!*b)
+		for( i=0; i<256; i++ )
+			BIT_ON( m->boundary, 0, i );
+
+	// Otherwise set the specified chars as boundary
+	else
+		for(; *b; b++ )
+			BIT_ON( m->boundary, 0, *b );
+			
+}
+
 int insert(SV* obj, SV* key, SV* val) {
 	fsm m = (fsm)SvIV(SvRV(obj));
 
@@ -639,8 +683,7 @@ int has(SV* obj, char *s) {
 	return _search(m->root, s);
 }
 
-// This handles keys() and values()
-void _traverse(SV* obj, bool isKeys){
+void dump(SV* obj){
 	fsm m = (fsm)SvIV(SvRV(obj));
 	int i;
 	SV** ptr;
@@ -648,17 +691,31 @@ void _traverse(SV* obj, bool isKeys){
 	INLINE_STACK_VARS;
 
 	k = (char*) malloc(sizeof(char) * m->maxpath);
-
 	_malloc(m);
-	if(isKeys){
-		_keys(m, m->root, k, 0);
-	}
-	else {
-		_values(m, m->root);
-	}
-	
+	_dump(m, m->root, k, 0);
+	free(k);
+
+	INLINE_STACK_RESET;
+    for (i = 0; i <= av_len(m->found_keys); i++) {
+		ptr = av_fetch(m->found_keys, i, 0);
+		INLINE_STACK_PUSH(sv_2mortal(newSVsv(*ptr)));
+		ptr = av_fetch(m->found_vals, i, 0);
+		INLINE_STACK_PUSH(sv_2mortal(newSVsv(*ptr)));
+    }
+    INLINE_STACK_DONE;
+}
+
+void keys(SV* obj){
+	fsm m = (fsm)SvIV(SvRV(obj));
+	int i;
+	SV** ptr;
+	char *k;
+	INLINE_STACK_VARS;
+
+	k = (char*) malloc(sizeof(char) * m->maxpath);
+	_malloc(m);
+	_dump(m, m->root, k, 0);
 	free(k);
-	/* now look at m->found_keys */
 
 	INLINE_STACK_RESET;
     for (i = 0; i <= av_len(m->found_keys); i++) {
@@ -666,30 +723,28 @@ void _traverse(SV* obj, bool isKeys){
 		INLINE_STACK_PUSH(sv_2mortal(newSVsv(*ptr)));
     }
     INLINE_STACK_DONE;
-
-}
-
-void keys(SV* obj){
-	_traverse(obj, TRUE);
 }
 
 void values(SV* obj){
-	_traverse(obj, FALSE);
-}
-
-
-// Deprecated, use states()
-int btrees(SV* obj){
 	fsm m = (fsm)SvIV(SvRV(obj));
-	return m->states;
-}
+	int i;
+	SV** ptr;
+	char *k;
+	INLINE_STACK_VARS;
 
-// Deprecated and inaccurate, use transitions()
-int nodes(SV* obj){
-	fsm m = (fsm)SvIV(SvRV(obj));
-	return m->transitions - m->terminals;
-}
+	k = (char*) malloc(sizeof(char) * m->maxpath);
+	_malloc(m);
+	_dump(m, m->root, k, 0);
+	free(k);
 
+	INLINE_STACK_RESET;
+    for (i = 0; i <= av_len(m->found_vals); i++) {
+		ptr = av_fetch(m->found_vals, i, 0);
+		INLINE_STACK_PUSH(sv_2mortal(newSVsv(*ptr)));
+    }
+    INLINE_STACK_DONE;
+
+}
 
 int states(SV* obj){
 	fsm m = (fsm)SvIV(SvRV(obj));
@@ -708,14 +763,14 @@ int terminals(SV* obj){
 
 
 
-void _relay(SV* obj, char *s, bool isMindex) {
+void scan(SV* obj, char *s) {
 	fsm m = (fsm)SvIV(SvRV(obj));
 	int i;
 	SV** ptr;
 	INLINE_STACK_VARS;
 	
 	_malloc(m);
-	_scan(m, m->root, s, isMindex);
+	_scan(m, s);
 	
 	INLINE_STACK_RESET;
 	for (i = 0; i <= av_len(m->found_keys); i++) {
@@ -725,25 +780,57 @@ void _relay(SV* obj, char *s, bool isMindex) {
 		INLINE_STACK_PUSH(sv_2mortal(newSVsv(*ptr)));
 	}
 	INLINE_STACK_DONE;
+}
 
+void mindex(SV* obj, char *s) {
+	fsm m = (fsm)SvIV(SvRV(obj));
+	int i;
+	SV** ptr;
+	INLINE_STACK_VARS;
+	
+	_malloc(m);
+	_scan(m, s);
+	
+	INLINE_STACK_RESET;
+	for (i = 0; i <= av_len(m->found_keys); i++) {
+		ptr = av_fetch(m->found_keys, i, 0);
+		INLINE_STACK_PUSH(sv_2mortal(newSVsv(*ptr)));
+		ptr = av_fetch(m->found_offsets, i, 0);
+		INLINE_STACK_PUSH(sv_2mortal(newSVsv(*ptr)));
+	}
+	INLINE_STACK_DONE;
+}
+
+void multiscan(SV* obj, char *s) {
+	fsm m = (fsm)SvIV(SvRV(obj));
+	int i;
+	SV** ptr;
+	AV* result; // holds one result (3-item array)
+	INLINE_STACK_VARS;
+	
+	_malloc(m);
+	_scan(m, s);
+	
+	INLINE_STACK_RESET;
+	for (i = 0; i <= av_len(m->found_keys); i++) {
+
+		result = (AV*) newAV();
+
+		ptr = av_fetch(m->found_keys, i, 0);
+		av_push(result, (SV*) newSVsv(*ptr));
+
+		ptr = av_fetch(m->found_offsets, i, 0);
+		av_push(result, (SV*) newSVsv(*ptr));
+
+		ptr = av_fetch(m->found_vals, i, 0);
+		av_push(result, (SV*) newSVsv(*ptr));
+
+		INLINE_STACK_PUSH(sv_2mortal(newRV_noinc((SV*)result)));
+	}
+	INLINE_STACK_DONE;
 }
 
 
-void mindex(SV* obj, char *s){
-	_relay(obj, s, TRUE);
-}
-
-void scan(SV* obj, char *s){
-	_relay(obj, s, FALSE);
-}
-
-
-// Vector records
-#define BIT_ON(vec, pos) \
-		*(vec+(int)pos/8) |= (1 << (pos % 8))
-
-#define IS_BIT_ON(vec, pos) \
-		*(vec+(int)pos/8) &  (1 << (pos % 8))
 
 // This inline uses variables in context
 #define RECORD_STATE \
@@ -798,7 +885,7 @@ int _serialize(SV* obj, char *triename, char *valsname){
 		// record state [and value if present] at front, 
 		// move front to end of state. Increment pos by len of state
 		RECORD_STATE;
-		BIT_ON(tvector, pos);
+		BIT_ON(tvector, 0, pos);
 		
 		if(!back){ break; } //the end
 		
@@ -814,7 +901,7 @@ int _serialize(SV* obj, char *triename, char *valsname){
 // Now repair the trie, severing horizontal links between states
 	front = m->root;
 	for( pos=0; pos < m->transitions; pos++ ){
-		if(IS_BIT_ON(tvector, pos)){
+		if(IS_BIT_ON(tvector, 0, pos)){
 			back = front;
 			front = front->next_trans;
 			back->next_trans = 0;
